"""
Test for app_settings.py
"""

# Django
from django.test import TestCase, modify_settings

# AA Forum
from aa_forum.app_settings import (
    aa_timezones_installed,
    allianceauth_discordbot_installed,
    discord_messaging_proxy_available,
)


class TestModulesInstalled(TestCase):
    """
    Test if modules are installed
    """

    @modify_settings(INSTALLED_APPS={"append": "aadiscordbot"})
    def test_allianceauth_discordbot_installed_should_return_true(self):
        """
        Test allianceauth_discordbot_installed should return True

        :return:
        :rtype:
        """

        self.assertTrue(expr=allianceauth_discordbot_installed())

    @modify_settings(INSTALLED_APPS={"remove": "aadiscordbot"})
    def test_allianceauth_discordbot_installed_should_return_false(self):
        """
        Test allianceauth_discordbot_installed should return False

        :return:
        :rtype:
        """

        self.assertFalse(expr=allianceauth_discordbot_installed())

    @modify_settings(INSTALLED_APPS={"append": "aadiscordbot"})
    def test_discord_messaging_proxy_available_return_true(self):
        """
        Test discord_messaging_proxy_available should return True for aadiscordbot

        :return:
        :rtype:
        """

        self.assertTrue(expr=discord_messaging_proxy_available())

    @modify_settings(INSTALLED_APPS={"append": "timezones"})
    def test_aa_timezones_installed_should_return_true(self):
        """
        Test aa_timezones_installed should return True

        :return:
        :rtype:
        """

        self.assertTrue(expr=aa_timezones_installed())

    @modify_settings(INSTALLED_APPS={"remove": "timezones"})
    def test_aa_timezones_installed_should_return_false(self):
        """
        Test aa_timezones_installed should return False

        :return:
        :rtype:
        """

        self.assertFalse(expr=aa_timezones_installed())



"""
Personal Messages Views
"""

# Standard Library
from http import HTTPStatus

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, JsonResponse
from django.shortcuts import redirect, render
from django.utils.datastructures import MultiValueDictKeyError
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Forum
from aa_forum import __title__
from aa_forum.forms import NewPersonalMessageForm, ReplyPersonalMessageForm
from aa_forum.helper.pagination import get_paginated_page_object
from aa_forum.helper.user import get_main_character_from_user
from aa_forum.models import PersonalMessage, Setting

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


@login_required
@permission_required(perm="aa_forum.basic_access")
def inbox(request: WSGIRequest, page_number: int = None) -> HttpResponse:
    """
    Overview of all messages in a user's inbox

    :param request:
    :type request:
    :param page_number:
    :type page_number:
    :return:
    :rtype:
    """

    logger.info(msg=f"{request.user} called their messages overview.")

    personal_messages = PersonalMessage.objects.get_personal_messages_for_user(
        user=request.user
    )

    page_obj = get_paginated_page_object(
        queryset=personal_messages,
        items_per_page=Setting.objects.get_setting(
            setting_key=Setting.Field.MESSAGESPERPAGE
        ),
        page_number=page_number,
    )

    context = {"page_obj": page_obj}

    return render(
        request=request,
        template_name="aa_forum/view/personal-messages/inbox.html",
        context=context,
    )


@login_required
@permission_required("aa_forum.basic_access")
def new_message(request: WSGIRequest) -> HttpResponse:
    """
    Create a new personal message

    :param request:
    :type request:
    :return:
    :rtype:
    """

    logger.info(msg=f"{request.user} called the new personal message page.")

    # If this is a POST request, we need to process the form data
    if request.method == "POST":
        new_private_message_form = NewPersonalMessageForm(data=request.POST)

        # Check whether it's valid:
        if new_private_message_form.is_valid():
            sender = request.user
            recipient = new_private_message_form.cleaned_data["recipient"]
            subject = new_private_message_form.cleaned_data["subject"]
            message = new_private_message_form.cleaned_data["message"]

            PersonalMessage(
                sender=sender,
                recipient=recipient,
                subject=subject,
                message=message,
            ).save()

            recipient_main_char = get_main_character_from_user(user=recipient)

            messages.success(
                request=request,
                message=mark_safe(
                    # pylint: disable=duplicate-code
                    s=_(
                        f"<h4>Success!</h4><p>Message to {recipient_main_char} sent.</p>"  # pylint: disable=line-too-long
                    )
                ),
            )

            return redirect(to="aa_forum:personal_messages_inbox")

        messages.error(
            request=request,
            message=mark_safe(
                # pylint: disable=duplicate-code
                s=_(
                    "<h4>Error!</h4>"
                    "<p>Something went wrong, please check your input.</p>"
                )
            ),
        )
    else:
        new_private_message_form = NewPersonalMessageForm()

    context = {"form": new_private_message_form}

    return render(
        request=request,
        template_name="aa_forum/view/personal-messages/new-message.html",
        context=context,
    )


@login_required
@permission_required(perm="aa_forum.basic_access")
def sent_messages(request: WSGIRequest, page_number: int = None) -> HttpResponse:
    """
    Overview of all messages in a users sent messages

    :param request:
    :type request:
    :param page_number:
    :type page_number:
    :return:
    :rtype:
    """

    logger.info(msg=f"{request.user} called the their sent personal message page.")

    personal_messages = PersonalMessage.objects.get_personal_messages_sent_for_user(
        user=request.user
    )

    page_obj = get_paginated_page_object(
        queryset=personal_messages,
        items_per_page=Setting.objects.get_setting(
            setting_key=Setting.Field.MESSAGESPERPAGE
        ),
        page_number=page_number,
    )

    context = {"page_obj": page_obj}

    return render(
        request=request,
        template_name="aa_forum/view/personal-messages/sent-messages.html",
        context=context,
    )


@login_required
@permission_required("aa_forum.basic_access")
def reply_message(request: WSGIRequest, message_id: int) -> HttpResponse:
    """
    Reply to a personal message

    :param request:
    :type request:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    context = {}

    try:
        personal_message = PersonalMessage.objects.get(
            pk=message_id, recipient=request.user, deleted_by_recipient=False
        )
    except PersonalMessage.DoesNotExist:
        messages.error(
            request=request,
            message=mark_safe(
                # pylint: disable=duplicate-code
                s=_(
                    "<h4>Error!</h4>"
                    "<p>The message you were trying to reply to does either not "
                    "exist or you are not the recipient.</p>"
                )
            ),
        )

        return redirect(to="aa_forum:personal_messages_inbox")

    # If this is a POST request, we need to process the form data
    if request.method == "POST":
        reply_private_message_form = ReplyPersonalMessageForm(data=request.POST)

        # Check whether it's valid:
        if reply_private_message_form.is_valid():
            sender = request.user
            recipient = personal_message.sender

            subject = personal_message.subject
            if not subject.startswith("Re:"):
                subject = "Re: " + subject

            message = reply_private_message_form.cleaned_data["message"]

            PersonalMessage(
                sender=sender,
                recipient=recipient,
                subject=subject,
                message=message,
            ).save()

            recipient_main_char = get_main_character_from_user(user=recipient)
            messages.success(
                request=request,
                message=mark_safe(
                    # pylint: disable=duplicate-code
                    s=_(f"<h4>Success!</h4><p>Reply to {recipient_main_char} sent.</p>")
                ),
            )

            return redirect(to="aa_forum:personal_messages_inbox")

        messages.error(
            request=request,
            message=mark_safe(
                # pylint: disable=duplicate-code
                s=_(
                    "<h4>Error!</h4>"
                    "<p>Something went wrong, please check your input.</p>"
                )
            ),
        )
    else:
        reply_private_message_form = ReplyPersonalMessageForm()

    context["message"] = personal_message
    context["form"] = reply_private_message_form

    return render(
        request=request,
        template_name="aa_forum/view/personal-messages/reply-message.html",
        context=context,
    )


@login_required
@permission_required("aa_forum.basic_access")
def delete_message(request: WSGIRequest, folder: str, message_id: int) -> HttpResponse:
    """
    Delete a personal message

    :param request:
    :type request:
    :param folder:
    :type folder:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    def folder_inbox() -> HttpResponse:
        """
        Remove message from inbox

        :return:
        """

        try:
            message = PersonalMessage.objects.get(pk=message_id, recipient=request.user)
        except PersonalMessage.DoesNotExist:
            messages.error(
                request=request,
                message=mark_safe(
                    # pylint: disable=duplicate-code
                    s=_(
                        "<h4>Error!</h4>"
                        "<p>The message you tried to remove does either not exist "
                        "or is not yours to remove.</p>"
                    )
                ),
            )
        else:
            if message.deleted_by_sender is True:
                message.delete()
            else:
                message.deleted_by_recipient = True
                message.save()

            messages.success(
                request=request,
                message=mark_safe(
                    s=_("<h4>Success!</h4><p>Message removed from your inbox.</p>")
                ),
            )

        return redirect(to="aa_forum:personal_messages_inbox")

    def folder_sent_messages() -> HttpResponse:
        """
        Remove a message from sent messages

        :return:
        """

        try:
            message = PersonalMessage.objects.get(pk=message_id, sender=request.user)
        except PersonalMessage.DoesNotExist:
            messages.error(
                request=request,
                message=mark_safe(
                    # pylint: disable=duplicate-code
                    s=_(
                        "<h4>Error!</h4>"
                        "<p>The message you tried to remove does either not exist "
                        "or is not yours to remove.</p>"
                    )
                ),
            )
        else:
            if message.deleted_by_recipient is True:
                message.delete()
            else:
                message.deleted_by_sender = True
                message.save()

            messages.success(
                request=request,
                message=mark_safe(
                    # pylint: disable=duplicate-code
                    s=_(
                        "<h4>Success!</h4>"
                        "<p>Message has been removed from your sent messages.</p>"
                    )
                ),
            )

        return redirect(to="aa_forum:personal_messages_sent_messages")

    switch = {"inbox": folder_inbox, "sent-messages": folder_sent_messages}

    if folder in switch:
        return switch[folder]()

    messages.error(
        request=request,
        message=mark_safe(s=_("<h4>Error!</h4><p>Something went wrong.</p>")),
    )

    return redirect(to="aa_forum:personal_messages_inbox")


@login_required
@permission_required(perm="aa_forum.basic_access")
def ajax_read_message(request: WSGIRequest, folder: str) -> HttpResponse:
    """
    Read a personal message

    :param request:
    :type request:
    :param folder:
    :type folder:
    :return:
    :rtype:
    """

    data = {}

    if request.method == "POST":
        try:
            sender_id = int(request.POST["sender"])
            recipient_id = int(request.POST["recipient"])
            message_id = int(request.POST["message"])
        except MultiValueDictKeyError:
            # Fail silently
            pass
        else:
            if (folder == "inbox" and request.user.id == recipient_id) or (
                folder == "sent-messages" and request.user.id == sender_id
            ):
                try:
                    message = PersonalMessage.objects.get(
                        pk=message_id, sender_id=sender_id, recipient_id=recipient_id
                    )
                except PersonalMessage.DoesNotExist:
                    # Fail silently
                    pass
                else:
                    # Mark message as read
                    if folder == "inbox" and message.is_read is False:
                        message.is_read = True
                        message.save()

                    data["message"] = message
                    data["folder"] = folder

                    return render(
                        request=request,
                        template_name="aa_forum/ajax-render/personal-messages/message.html",  # pylint: disable=line-too-long
                        context=data,
                    )

    return HttpResponse(status=HTTPStatus.NO_CONTENT)


@login_required
@permission_required("aa_forum.basic_access")
def ajax_unread_messages_count(request: WSGIRequest) -> JsonResponse:
    """
    Get the number of unread messages for a user

    :param request:
    :type request:
    :return:
    :rtype:
    """

    unread_messages_count = (
        PersonalMessage.objects.get_personal_message_unread_count_for_user(
            user=request.user
        )
    )

    data = {"unread_messages_count": unread_messages_count}

    return JsonResponse(data=data, safe=False)



"""
Test managers
"""

# Django
from django.contrib.auth.models import Group
from django.test import TestCase

# AA Forum
from aa_forum.models import Board, Category, Topic
from aa_forum.tests.utils import create_fake_message, create_fake_user


class TestBoard(TestCase):
    """
    Tests for board manager
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and categories

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")
        cls.category = Category.objects.create(name="Science")

    def setUp(self) -> None:
        """
        Set up user

        :return:
        :rtype:
        """

        self.user = create_fake_user(character_id=1001, character_name="Bruce Wayne")

    def test_should_return_board_with_no_groups(self):
        """
        Test should return board with no groups

        :return:
        :rtype:
        """

        # given
        board = Board.objects.create(name="Physics", category=self.category)

        # when
        result = Board.objects.user_has_access(user=self.user)

        # then
        self.assertIn(member=board, container=result)

    def test_should_return_board_for_group_member(self):
        """
        Test should return board for group member

        :return:
        :rtype:
        """

        # given
        board = Board.objects.create(name="Physics", category=self.category)
        board.groups.add(self.group)
        self.user.groups.add(self.group)

        # when
        result = Board.objects.user_has_access(user=self.user)

        # then
        self.assertIn(member=board, container=result)

    def test_should_not_return_board_for_non_group_member(self):
        """
        Test should not return board for non group member

        :return:
        :rtype:
        """

        # given
        board = Board.objects.create(name="Physics", category=self.category)
        board.groups.add(self.group)

        # when
        result = Board.objects.user_has_access(user=self.user)

        # then
        self.assertNotIn(member=board, container=result)


class TestTopic(TestCase):
    """
    Tests for topic manager
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and categories

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")
        cls.category = Category.objects.create(name="Science")

    def setUp(self) -> None:
        """
        Set up user and board

        :return:
        :rtype:
        """

        self.user = create_fake_user(character_id=1001, character_name="Bruce Wayne")
        self.board = Board.objects.create(name="Physics", category=self.category)

    def test_should_return_topic_normally_1(self):
        """
        Test should return a topic normally (test with keyword arguments)

        :return:
        :rtype:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic=topic, user=self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug=str(self.category.slug),
            board_slug=str(self.board.slug),
            topic_slug=str(topic.slug),
            user=self.user,
        )

        # then
        self.assertEqual(first=result, second=topic)

    def test_should_return_topic_normally_2(self):
        """
        Test should return a topic normally (test with positional arguments)

        :return:
        :rtype:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic=topic, user=self.user)

        # when
        result = Topic.objects.get_from_slugs(
            self.category.slug,
            self.board.slug,
            topic.slug,
            self.user,
        )

        # then
        self.assertEqual(first=result, second=topic)

    def test_should_return_none_if_not_found_1(self):
        """
        Test should return None if not found (invalid topic slug)

        :return:
        :rtype:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic=topic, user=self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug=self.category.slug,
            board_slug=self.board.slug,
            topic_slug="invalid",
            user=self.user,
        )

        # then
        self.assertIsNone(obj=result)

    def test_should_return_none_if_not_found_2(self):
        """
        Test should return None if not found (invalid board slug)

        :return:
        :rtype:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic=topic, user=self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug=self.category.slug,
            board_slug="invalid",
            topic_slug=topic.slug,
            user=self.user,
        )

        # then
        self.assertIsNone(obj=result)

    def test_should_return_none_if_not_found_3(self):
        """
        Test should return None if not found (invalid category slug)

        :return:
        :rtype:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic=topic, user=self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug="invalid",
            board_slug=self.board.slug,
            topic_slug=topic.slug,
            user=self.user,
        )

        # then
        self.assertIsNone(obj=result)

    def test_should_return_restricted_topic_for_group_members(self):
        """
        Test should return a restricted topic for group members

        :return:
        :rtype:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        self.user.groups.add(self.group)
        self.board.groups.add(self.group)
        create_fake_message(topic=topic, user=self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug=self.category.slug,
            board_slug=self.board.slug,
            topic_slug=topic.slug,
            user=self.user,
        )

        # then
        self.assertEqual(first=result, second=topic)

    def test_should_not_return_restricted_topic_for_non_group_members(self):
        """
        Test should not return a restricted topic for non group members

        :return:
        :rtype:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        self.board.groups.add(self.group)
        create_fake_message(topic=topic, user=self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug=self.category.slug,
            board_slug=self.board.slug,
            topic_slug=topic.slug,
            user=self.user,
        )

        # then
        self.assertIsNone(obj=result)
